/*
 * Copyright (C) 2005 Red Hat, Inc.
 *
 * This file is released under the LGPL.
 */

#include <linux/init.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/vmalloc.h>
#include <linux/list.h>
#include <linux/net.h>
#include <linux/in.h>
#include <linux/socket.h>
#include <linux/signal.h>
#include <linux/mempool.h>
#include <linux/bio.h>
#include <linux/blkdev.h>
#include <linux/device-mapper.h>
#include <linux/delay.h>
#include <cluster/service.h>
#include <cluster/cnxman.h>
#include <cluster/cnxman-socket.h>

#include "dm-log.h"
#include "dm-cmirror-xfr.h"
#include "dm-cmirror-common.h"
#include "dm-cmirror-server.h"
#include "dm-cmirror-cman.h"

DECLARE_MUTEX(log_list_lock);
LIST_HEAD(log_list_head);

struct region_state {
	int rs_mark_logged;
	region_t rs_region;
	struct list_head rs_list;
};

static mempool_t *region_state_pool = NULL;

static int shutting_down=0;
static atomic_t suspend_client;
static wait_queue_head_t suspend_client_queue;
static wait_queue_head_t event_queue;

static DECLARE_MUTEX(consult_server_lock);

/* These vars are just for stats, and will be removed */
static uint32_t request_count=0;
static uint32_t request_retry_count=0;

static void *region_state_alloc(int gfp_mask, void *pool_data){
	return kmalloc(sizeof(struct region_state), gfp_mask);
}

static void region_state_free(void *element, void *pool_data){
	kfree(element);
}

#define BYTE_SHIFT 3
/*
 *   <region_size> <uuid> [[no]sync] [block_on_error]
 */
static int core_ctr(struct dirty_log *log, struct dm_target *ti,
		    unsigned int argc, char **argv)
{
	enum sync sync = DEFAULTSYNC;
	int failure_response = FR_NONBLOCK;

	struct log_c *lc;
	sector_t region_size;
	unsigned int region_count;
	size_t bitset_size;
	int uuid = 0;
	int i;

	/* Already checked argument count */

	for (i = 1; i < argc; i++) {
		if (!strcmp(argv[i], "sync"))
			sync = FORCESYNC;
		else if (!strcmp(argv[i], "nosync"))
			sync = NOSYNC;
		else if (!strcmp(argv[i], "block_on_error"))
			failure_response = FR_BLOCK;
		else if (!uuid)
			uuid = i;
		else {
			DMWARN("unrecognised argument to clustered mirror log: %s",
			       argv[i]);
			return -EINVAL;
		}
	}

	if (sscanf(argv[0], SECTOR_FORMAT, &region_size) != 1) {
		DMWARN("invalid region size string");
		return -EINVAL;
	}

	region_count = dm_sector_div_up(ti->len, region_size);

	lc = kmalloc(sizeof(*lc), GFP_KERNEL);
	if (!lc) {
		DMWARN("Couldn't allocate core log");
		return -ENOMEM;
	}
	memset(lc, 0, sizeof(*lc));

	lc->ti = ti;
	lc->region_size = region_size;
	lc->region_count = region_count;
	lc->sync = sync;
	lc->failure_response = failure_response;
	strncpy(lc->uuid, argv[uuid], MAX_NAME_LEN);
	init_completion(&lc->complete);

	/*
	 * Work out how many words we need to hold the bitset.
	 */
	bitset_size = dm_round_up(region_count,
				  sizeof(*lc->clean_bits) << BYTE_SHIFT);

	bitset_size >>= BYTE_SHIFT;

	lc->bitset_uint32_count = bitset_size / sizeof(*lc->clean_bits);
	lc->clean_bits = vmalloc(bitset_size);
	if (!lc->clean_bits) {
		DMWARN("couldn't allocate clean bitset");
		kfree(lc);
		return -ENOMEM;
	}
	memset(lc->clean_bits, -1, bitset_size);

	lc->sync_bits = vmalloc(bitset_size);
	if (!lc->sync_bits) {
		DMWARN("couldn't allocate sync bitset");
		vfree(lc->clean_bits);
		kfree(lc);
		return -ENOMEM;
	}
	memset(lc->sync_bits, (sync == NOSYNC) ? -1 : 0, bitset_size);
	lc->sync_count = (sync == NOSYNC) ? region_count : 0;

	lc->recovering_region = (uint64_t)-1;
	lc->recovering_next = (uint64_t)-1;
	lc->sync_search = 0;
	log->context = lc;
	return 0;
}

static void core_dtr(struct dirty_log *log)
{
	struct log_c *lc = (struct log_c *) log->context;
	vfree(lc->clean_bits);
	vfree(lc->sync_bits);
	kfree(lc);
}

/*----------------------------------------------------------------
 * disk log constructor/destructor
 *
 *   <device> <region_size> <uuid> [[no]sync] [block_on_error]
 *--------------------------------------------------------------*/
static int disk_ctr(struct dirty_log *log, struct dm_target *ti,
		    unsigned int argc, char **argv)
{
	int r;
	size_t size;
	struct log_c *lc;
	struct dm_dev *dev;

	/* already checked argument count */
	r = dm_get_device(ti, argv[0], 0, 0 /* FIXME */,
			  FMODE_READ | FMODE_WRITE, &dev);
	if (r){
		DMWARN("Unable to get device %s", argv[0]);
		return r;
	}

	r = core_ctr(log, ti, argc - 1, argv + 1);
	if (r) {
		dm_put_device(ti, dev);
		return r;
	}

	lc = (struct log_c *) log->context;
	lc->log_dev = dev;
	lc->log_dev_failed = 0;
	/* init_completion(&lc->failure_completion); */

	/* setup the disk header fields */
	lc->header_location.bdev = lc->log_dev->bdev;
	lc->header_location.sector = 0;
	lc->header_location.count = 1;

	/*
	 * We can't read less than this amount, even though we'll
	 * not be using most of this space.
	 */
	lc->disk_header = vmalloc(1 << SECTOR_SHIFT);
	if (!lc->disk_header)
		goto bad;

	/* setup the disk bitset fields */
	lc->bits_location.bdev = lc->log_dev->bdev;
	lc->bits_location.sector = LOG_OFFSET;

	size = dm_round_up(lc->bitset_uint32_count * sizeof(uint32_t),
			   1 << SECTOR_SHIFT);
	lc->bits_location.count = size >> SECTOR_SHIFT;

	return 0;

 bad:
	dm_put_device(ti, lc->log_dev);
	core_dtr(log);
	return -ENOMEM;
}

static void disk_dtr(struct dirty_log *log)
{
	struct log_c *lc = (struct log_c *) log->context;
	dm_put_device(lc->ti, lc->log_dev);
	vfree(lc->disk_header);
	core_dtr(log);
}



static int run_election(struct log_c *lc, uint32_t initial_server){
	int error=0, len;
	struct sockaddr_in saddr_in;
	struct msghdr msg;
	struct iovec iov;
	mm_segment_t fs;
	struct log_request lr;  /* ATTENTION -- could be too much on the stack */
  
	memset(&lr, 0, sizeof(lr));

	lr.lr_type = LRT_ELECTION;
	lr.u.lr_starter = my_id;
	lr.u.lr_coordinator = initial_server;
	memcpy(lr.lr_uuid, lc->uuid, MAX_NAME_LEN);
	lr.lr_uuid_ref = lc->uuid_ref;

	memset(&saddr_in, 0, sizeof(struct sockaddr_cl));

	msg.msg_control = NULL;
	msg.msg_controllen = 0;
	msg.msg_iovlen = 1;
	msg.msg_iov = &iov;
	msg.msg_flags = 0;
  
	saddr_in.sin_family = AF_INET;
	saddr_in.sin_port = CLUSTER_LOG_PORT;
	if(!(saddr_in.sin_addr.s_addr = nodeid_to_ipaddr(my_id))){
		DMERR("Unable to convert nodeid_to_ipaddr in run_election");
	}
	msg.msg_name = &saddr_in;
	msg.msg_namelen = sizeof(saddr_in);

	iov.iov_len = sizeof(struct log_request);
	iov.iov_base = &lr;

	fs = get_fs();
	set_fs(get_ds());

	len = sock_sendmsg(lc->client_sock, &msg, sizeof(struct log_request));

	if(len < 0){
		DMERR("unable to send election notice to server (error = %d)", len);
		error = len;
		set_fs(fs);
		goto fail;
	}

  
	/* why do we need to reset this? */
	iov.iov_len = sizeof(struct log_request);
	iov.iov_base = &lr;

	len = my_recvmsg(lc->client_sock, &msg, sizeof(struct log_request),
			 0, 20);
	set_fs(fs);
  
	if(len > 0){
		lc->server_id = lr.u.lr_coordinator;
	} else {
		/* ATTENTION -- what do we do with this ? */
		DMWARN("Failed to receive election results from server: (%s/%d,%d)",
		       lc->uuid + (strlen(lc->uuid) - 8), lc->uuid_ref, len);
		error = len;
	}

 fail:
	return error;
}

static int _consult_server(struct log_c *lc, region_t region,
			  int type, region_t *result, int *retry){
	static int seq = 0;
	int len;
	int error=0;
	struct sockaddr_in saddr_in;
	struct msghdr msg;
	struct iovec iov;
	mm_segment_t fs;
	struct log_request *lr;

	request_count++;

	lr = kmalloc(sizeof(struct log_request), GFP_NOFS);
	if (!lr) {
		BUG();
		error = -ENOMEM;
		goto retry;
	}

	memset(lr, 0, sizeof(struct log_request));
	
	lr->lr_type = type;
	lr->lr_seq = seq;
	if(type == LRT_MASTER_LEAVING){
		lr->u.lr_starter = my_id;
	} else {
		lr->u.lr_region = region;
	}

	if (type == LRT_COMPLETE_RESYNC_WORK)
		lr->u.lr_int_rtn = (*result) ? 1 : 0;

	memcpy(lr->lr_uuid, lc->uuid, MAX_NAME_LEN);
	lr->lr_uuid_ref = lc->uuid_ref;

	memset(&saddr_in, 0, sizeof(struct sockaddr_in));

	msg.msg_control = NULL;
	msg.msg_controllen = 0;
	msg.msg_iovlen = 1;
	msg.msg_iov = &iov;
	msg.msg_flags = 0;
  
	saddr_in.sin_family = AF_INET;
	saddr_in.sin_port = CLUSTER_LOG_PORT;
	if (!(saddr_in.sin_addr.s_addr = nodeid_to_ipaddr(lc->server_id))) {
		error = -ENXIO;
		goto retry;
	}
	msg.msg_name = &saddr_in;
	msg.msg_namelen = sizeof(saddr_in);

	iov.iov_len = sizeof(struct log_request);
	iov.iov_base = lr;

	fs = get_fs();
	set_fs(get_ds());
  
	len = sock_sendmsg(lc->client_sock, &msg, sizeof(struct log_request));

	set_fs(fs);

	if (len < sizeof(struct log_request)) {
		DMWARN("unable to send log request to server");
		error = -EBADE;
		goto fail;
	}

rerecv:
	iov.iov_len = sizeof(struct log_request);
	iov.iov_base = lr;

	fs = get_fs();
	set_fs(get_ds());

	len = my_recvmsg(lc->client_sock, &msg,
			 sizeof(struct log_request), 0, 15);
	set_fs(fs);

	if (len <= 0) {
		/* ATTENTION -- what do we do with this ? */
		DMWARN("Error listening for server(%u) response for %s: %d",
		       lc->server_id, lc->uuid + (strlen(lc->uuid) - 8), len);
		error = len;
		seq++;
		goto retry;
	}
    
	if (seq != lr->lr_seq) {
		DMDEBUG("Message sequence number mismatch: %d/%d",
		      seq, lr->lr_seq);
		if (seq > lr->lr_seq) {
			DMDEBUG(" Skipping.  Listening again for response to %s",
			      RQ_STRING(type));
			memset(lr, 0, sizeof(struct log_request));
			goto rerecv;
		}
		DMERR(" Seq# mismatch: Must try to resend request, %s", RQ_STRING(type));
		error = -EBADE;
		seq++;
		goto retry;
	}
	seq++;

	if (type != lr->lr_type) {
		DMERR("Got incorrect message type back: %s/%s",
		      RQ_STRING(type), RQ_STRING(lr->lr_type));
		error = -EBADE;
		goto retry;
	}

	if (memcmp(lc->uuid, lr->lr_uuid, MAX_NAME_LEN)) {
		DMERR("Got reply from server for wrong log:");
		DMERR(" Expected UUID: %s", lc->uuid);
		DMERR(" Recieved UUID: %s", lr->lr_uuid);
		error = -EBADE;
		goto retry;
	}

	if (lr->u.lr_int_rtn == -EAGAIN) {
		DMWARN("Server (%u), request type %d, -EAGAIN.",
		       lc->server_id, lr->lr_type);
		goto retry;
	}

	if (lr->u.lr_int_rtn == -ENXIO) {
		DMDEBUG("Server (%u) says it no longer controls this log (%s)",
			lc->server_id, lc->uuid + (strlen(lc->uuid) - 8));
		lc->server_id = 0xDEAD;
		goto retry;
	}

	if (result)
		*result = lr->u.lr_region_rtn;

	error = lr->u.lr_int_rtn;
	kfree(lr);
	return error;

 retry:
	*retry = 1;
	request_retry_count++;
	if (!(request_retry_count & 0x1F)) {
		DMINFO("Clustered mirror retried requests :: %u of %u (%u%%)",
		       request_retry_count,
		       request_count,
		       dm_div_up(request_retry_count*100, request_count));
		DMDEBUG("Last request:");
		DMDEBUG(" - my_id   :: %u", my_id);
		DMDEBUG(" - server  :: %u", lc->server_id);
		DMDEBUG(" - log uuid:: %s (%s)",
			lc->uuid + (strlen(lc->uuid) - 8),
			atomic_read(&lc->suspended) ? "suspended" : "active");
		DMDEBUG(" - request :: %s", RQ_STRING(type));
		DMDEBUG(" - error   :: %d", error);
		DMINFO("Too many retries, attempting to re-establish server connection.");
		lc->server_id = 0xDEAD;
	}

fail:
	if (lr)
		kfree(lr);
	return error;
}

static int consult_server(struct log_c *lc, region_t region,
			  int type, region_t *result){
	int rtn=0;
	int retry=0;
	int new_server=0;
	struct region_state *rs=NULL;

	/* ATTENTION -- need to change this, the server could fail at anypoint **
	** we do not want to send requests to the wrong place, or fail to run  **
	** an election when needed */
	down(&consult_server_lock);

	do{
		retry = 0;
		suspend_on(&suspend_client_queue, atomic_read(&suspend_client));
		if ((type == LRT_MASTER_LEAVING) && (lc->server_id == 0xDEAD)) {
			/* Nothing to do */
			goto out;
		}
	election:
		while ((lc->server_id == 0xDEAD) || (!lc->server_valid)) {
			DMDEBUG("server_id=%x, server_valid=%u, %s",
				lc->server_id, lc->server_valid,
				lc->uuid + (strlen(lc->uuid) - 8));
			DMDEBUG("trigger = %s", RQ_STRING(type));
			run_election(lc, my_id);
			new_server = 1;
			lc->server_valid = 1;
		}

		spin_lock(&lc->state_lock);
		if(new_server && 
		   !list_empty(&lc->mark_logged)){
			int i=0;
			LIST_HEAD(mark);

			DMINFO("Clean-up required due to new server");
			DMINFO(" - Resending all mark region requests");
			list_splice_init(&lc->mark_logged, &mark);

			spin_unlock(&lc->state_lock);

			list_for_each_entry(rs, &mark, rs_list){
				do {
					retry = 0;
					rtn = _consult_server(lc, rs->rs_region,
							      LRT_MARK_REGION, NULL, &retry);
					if (lc->server_id == 0xDEAD) {
						goto election;
					}
				} while(retry);
				i++;
			}

			spin_lock(&lc->state_lock);
			list_splice_init(&mark, &lc->mark_logged);

			DMINFO(" - %d mark region requests resent", i);
			DMINFO("Clean-up complete");
			DMINFO("Continuing request type, %d (%s)", type,
			       RQ_STRING(type));
			new_server = 0;
		}
		spin_unlock(&lc->state_lock);
		
		retry = 0;
		rtn = _consult_server(lc, region, type, result, &retry);
	} while(retry);
out:
	up(&consult_server_lock);

	return rtn;
}


static int cluster_connect(void);
static int cluster_disconnect(void);

static int cluster_ctr(struct dirty_log *log, struct dm_target *ti,
		       unsigned int argc, char **argv, int disk)
{
	int error = 0;
	struct log_c *lc, *tmp_lc;
	struct sockaddr_in saddr_in;

	if (!disk) {
		if ((error = core_ctr(log, ti, argc, argv))) {
			DMWARN("Clustered mirror:: core_ctr failed");
			return error;
		}
	} else {
		/* NOTE -- we take advantage of the fact that disk_ctr does **
		** not actually read the disk.  I suppose, however, that if **
		** it does in the future, we will simply reread it when a   **
		** server is started here.................................. */

		if((error = disk_ctr(log, ti, argc, argv))) {
			DMWARN("Clustered mirror:: disk_ctr failed");
			return error;
		}
	}

	lc = log->context;

	/* Pull this out to support pvmove
	if (lc->failure_response != FR_BLOCK) {
		DMERR("Clustered mirror requires \"block_on_error\" parameter");
		error = -EINVAL;
		goto fail;
	}
	*/

	atomic_set(&lc->in_sync, -1);
	lc->uuid_ref = 1;

	down(&log_list_lock);
	list_for_each_entry(tmp_lc, &log_list_head, log_list){
		if(!strncmp(tmp_lc->uuid, lc->uuid, MAX_NAME_LEN)){
			lc->uuid_ref = (lc->uuid_ref > tmp_lc->uuid_ref) ?
				lc->uuid_ref : tmp_lc->uuid_ref + 1;
			/*
			 * A second instance of the same log
			 * should start out suspended.
			 */
			atomic_set(&lc->suspended, 1);
			lc->recovery_halted = 1;
			DMDEBUG("Secondary log... suspended and recovery_halted");
		}
	}

	list_add(&lc->log_list, &log_list_head);
	up(&log_list_lock);
	DMDEBUG("Creating %s (%d)",
	       lc->uuid + (strlen(lc->uuid) - 8),
	       lc->uuid_ref);

	INIT_LIST_HEAD(&lc->region_users);
	INIT_LIST_HEAD(&lc->clear_waiting);
	INIT_LIST_HEAD(&lc->mark_waiting);
	INIT_LIST_HEAD(&lc->mark_logged);
	spin_lock_init(&lc->state_lock);

	lc->server_valid = 0;
	lc->server_id = 0xDEAD;

	if ((error = cluster_connect())) {
		DMWARN("Unable to connect to cluster infrastructure.");
		goto fail;
	}

	error = sock_create(AF_INET, SOCK_DGRAM,
			    0,
			    &lc->client_sock);

	if(error){
		DMWARN("unable to create clustered log client socket");
		goto fail;
	}

	saddr_in.sin_family = AF_INET;
	saddr_in.sin_port = CLUSTER_LOG_PORT+1;
	if(!(saddr_in.sin_addr.s_addr = nodeid_to_ipaddr(my_id))){
		DMERR("Unable to convert nodeid_to_ipaddr in cluster_ctr");
	}
	error = lc->client_sock->ops->bind(lc->client_sock,
					   (struct sockaddr *)&saddr_in,
					   sizeof(struct sockaddr_in));
	while(error == -EADDRINUSE){
		saddr_in.sin_port++;
		error = lc->client_sock->ops->bind(lc->client_sock,
						   (struct sockaddr *)&saddr_in,
						   sizeof(struct sockaddr_in));
	}

	if(error){
		DMWARN("unable to bind clustered log client socket");
		sock_release(lc->client_sock);
		goto fail;
	}

	return 0;

 fail:
	if (lc->log_dev)
		disk_dtr(log);
	else
		core_dtr(log);

	return error;
}

/*------------------------------------------------------------------
 * clustered_core log constructor
 * (preceding args::  <start> <len> mirror clustered_core <log_args>
 *
 * Right now, 2 <= argc <= 4.
 *
 * argv contains:
 *   <region_size> <uuid> [[no]sync] [block_on_error]
 *--------------------------------------------------------------*/
static int cluster_core_ctr(struct dirty_log *log, struct dm_target *ti,
		       unsigned int argc, char **argv) {
	int i;
	if ((argc < 2) || (argc > 4)) {
		DMERR("Too %s arguments to clustered_core mirror log type.",
		      (argc < 2) ? "few" : "many");
		DMERR("  %d arguments supplied:", argc);
		for (i = 0; i < argc; i++)
			DMERR("    %s", argv[i]);
		return -EINVAL;
	}

	return cluster_ctr(log, ti, argc, argv, 0);
}

/*------------------------------------------------------------------
 * clustered_disk log constructor
 * (preceding args::  <start> <len> mirror clustered_disk <log_args>
 *
 * Right now, 3 <= argc <= 5.
 *
 * argv contains:
 *   <disk> <region_size> <uuid> [[no]sync] [block_on_error]
 *--------------------------------------------------------------*/
static int cluster_disk_ctr(struct dirty_log *log, struct dm_target *ti,
		       unsigned int argc, char **argv) {
	int i;
	if ((argc < 3) || (argc > 5)) {
		DMERR("Too %s arguments to clustered_disk mirror log type.",
		      (argc < 3) ? "few" : "many");
		DMERR("  %d arguments supplied:", argc);
		for (i = 0; i < argc; i++)
			DMERR("    %s", argv[i]);
		return -EINVAL;
	}

	return cluster_ctr(log, ti, argc, argv, 1);
}

static void cluster_dtr(struct dirty_log *log)
{
	struct log_c *lc = (struct log_c *) log->context;
	struct region_state *rs, *tmp_rs;

	DMDEBUG("Removing %s (%d)",
	       lc->uuid + (strlen(lc->uuid) - 8),
	       lc->uuid_ref);

	down(&log_list_lock);
	list_del_init(&lc->log_list);
	up(&log_list_lock);

	if ((lc->server_id == my_id) && !atomic_read(&lc->suspended))
		consult_server(lc, 0, LRT_MASTER_LEAVING, NULL);

	sock_release(lc->client_sock);

	spin_lock(&lc->state_lock);

	if (!list_empty(&lc->clear_waiting)) {
		DMINFO("Clear requests remain at cluster log deactivation");
		list_for_each_entry_safe(rs, tmp_rs, &lc->clear_waiting, rs_list) {
			list_del_init(&rs->rs_list);
			DMINFO(" - Ignoring clear request: %Lu", rs->rs_region);
			mempool_free(rs, region_state_pool);
		}
	}

	if (!list_empty(&lc->mark_waiting)) {
		DMERR("Pending mark requests remain at cluster_dtr");
		BUG();
	}

	if (!list_empty(&lc->mark_logged)) {
		int i = 0;
		
		list_for_each_entry_safe(rs, tmp_rs, &lc->mark_logged, rs_list) {
			i++;
			list_del_init(&rs->rs_list);
			mempool_free(rs, region_state_pool);
		}
		DMDEBUG("%d mark requests remain at cluster log deactivation", i);
	}

	spin_unlock(&lc->state_lock);

	server_free_region_users(lc);

	if (lc->log_dev)
		disk_dtr(log);
	else
		core_dtr(log);

	if (cluster_disconnect())
		DMERR("Unable to disconnect from cluster infrastructure.\n");
}

static int cluster_presuspend(struct dirty_log *log)
{
	struct log_c *lc = (struct log_c *) log->context;

	DMDEBUG("cluster_presuspend: recovery halted on %s(%d)",
		lc->uuid + (strlen(lc->uuid) - 8), lc->uuid_ref);
	lc->recovery_halted = 1;
	return 0;
}

static int cluster_postsuspend(struct dirty_log *log)
{
	int i;
	struct region_state *rs, *tmp_rs;
	struct log_c *lc = (struct log_c *) log->context;

	DMDEBUG("cluster_postsuspend");
	spin_lock(&lc->state_lock);
	if (!list_empty(&lc->mark_waiting)) {
		DMERR("Mark requests remain at postsuspend!");
		BUG();
	}

	if (!list_empty(&lc->clear_waiting)) {
		DMERR("Clear requests remain at postsuspend!");

		list_for_each_entry_safe(rs, tmp_rs, &lc->clear_waiting, rs_list) {
			list_del_init(&rs->rs_list);
			DMERR(" - Ignoring clear request: %Lu", rs->rs_region);
			mempool_free(rs, region_state_pool);
		}
	}

	spin_unlock(&lc->state_lock);

	if(lc->server_id == my_id) {
		for (i = 0; server_busy(lc) && (i < 10); i++) {
			DMDEBUG("Server for %s still busy, waiting for others",
				lc->uuid + (strlen(lc->uuid) - 8));
			set_current_state(TASK_INTERRUPTIBLE);
			schedule_timeout(HZ*2);
		}
	}

	atomic_set(&lc->suspended, 1);
	if(lc->server_id == my_id) {
		while (1) {
			DMDEBUG("Telling everyone I'm suspending (%s)",
				lc->uuid + (strlen(lc->uuid) - 8));
			consult_server(lc, 0, LRT_MASTER_LEAVING, NULL);

			down(&consult_server_lock);
			run_election(lc, 0xDEAD);
			up(&consult_server_lock);

			if ((my_id && (lc->server_id == my_id))) {
				DMDEBUG("Delaying suspend, work to be done (%s)",
					lc->uuid + (strlen(lc->uuid) - 8));
				atomic_set(&lc->suspended, 0);
				set_current_state(TASK_INTERRUPTIBLE);
				schedule_timeout(HZ*2);
				atomic_set(&lc->suspended, 1);
			} else {
				DMDEBUG("Suspending now (%s)",
					lc->uuid + (strlen(lc->uuid) - 8));
				break;
			}
		}
	}
	atomic_set(&lc->suspended, 2);

	return 0;
}

static int cluster_resume(struct dirty_log *log){
	struct log_c *lc = (struct log_c *) log->context;

	lc->sync_search = 0;
	resume_server_requests();
	DMDEBUG("cluster_resume: Setting recovery_halted = 0");
	lc->recovery_halted = 0;
	atomic_set(&lc->suspended, 0);
	atomic_set(&lc->in_sync,0);

	return 0;
}

static uint32_t cluster_get_region_size(struct dirty_log *log)
{
	struct log_c *lc = (struct log_c *) log->context;

	return lc->region_size;
}


static int cluster_is_clean(struct dirty_log *log, region_t region)
{
	int rtn;
	struct log_c *lc = (struct log_c *) log->context;

	rtn = consult_server(lc, region, LRT_IS_CLEAN, NULL);

	return rtn;
}

static int cluster_is_remote_recovering(struct dirty_log *log, region_t region)
{
	int rtn;
	struct log_c *lc = (struct log_c *) log->context;
	 
	if (atomic_read(&lc->in_sync) == 1) {
		return 0;
	}

	rtn = consult_server(lc, region, LRT_IS_REMOTE_RECOVERING, NULL);
	return rtn;
}

static int cluster_in_sync(struct dirty_log *log, region_t region, int block)
{
	int rtn;
	struct log_c *lc = (struct log_c *) log->context;
  
	/* check known_regions, return if found */

	if(atomic_read(&lc->in_sync) == 1){
		return 1;
	}

	if(!block){
		return -EWOULDBLOCK;
	}

	rtn = consult_server(lc, region, LRT_IN_SYNC, NULL);

	return rtn;
}

static int cluster_flush(struct dirty_log *log)
{
	int r = 0;
	int clear_count = 0;
	int mark_count = 0;
	struct log_c *lc = (struct log_c *) log->context;
	struct region_state *rs, *tmp_rs;
	LIST_HEAD(mark);
	LIST_HEAD(clear);

	/*
	 * It should never be a problem to temporarily have
	 * the mark requests in limbo.  The only functions
	 * that call cluster_flush are rh_update_states and
	 * do_writes, and they are in the same thread as
	 * those changing the region states
	 */
	spin_lock(&lc->state_lock);
	list_splice_init(&lc->clear_waiting, &clear);
	list_splice_init(&lc->mark_waiting, &mark);
	spin_unlock(&lc->state_lock);

	list_for_each_entry_safe(rs, tmp_rs, &clear, rs_list) {
		/* don't really care if LRT_CLEAR_REGION fails */
		consult_server(lc, rs->rs_region, LRT_CLEAR_REGION, NULL);
		list_del_init(&rs->rs_list);
		mempool_free(rs, region_state_pool);
		clear_count++;
	}

	list_for_each_entry_safe(rs, tmp_rs, &mark, rs_list) {
		while (1) {
			r = consult_server(lc, rs->rs_region,
					   LRT_MARK_REGION, NULL);
			if (!r)
				break;

			if (r == -EBUSY) {
				DMDEBUG("Delaying mark to region %Lu, due to recovery",
					rs->rs_region);
				set_current_state(TASK_INTERRUPTIBLE);
				schedule_timeout(HZ);
				continue;
			}

			if (r == -EIO)
				goto fail;

			DMWARN("unable to get server (%u) to mark region (%Lu)",
			       lc->server_id, rs->rs_region);
			DMWARN("Reason :: %d", r);
		}
		mark_count++;
	}

	/* No flush work? */
	if (!clear_count && !mark_count)
		return 0;

	spin_lock(&lc->state_lock);
	list_splice_init(&mark, &lc->mark_logged);
	spin_unlock(&lc->state_lock);

	while ((r = consult_server(lc, 0, LRT_FLUSH, NULL))) {
		if (r == -EBUSY) {
			DMDEBUG("Delaying flush due to recovery (%s)",
				lc->uuid + (strlen(lc->uuid) - 8));
			set_current_state(TASK_INTERRUPTIBLE);
			schedule_timeout(HZ);
			continue;
		}

		if (r == -EIO)
			break;
	}

fail:
	if (r) {
		DMERR("Log flush failure: %d%s", r,
		      (r == -EIO) ? " -EIO" : "");
		if (lc->failure_response == FR_BLOCK)
			dm_table_event(lc->ti->table);
		lc->log_dev_failed = 1;
	}

	return r;
}

static void cluster_mark_region(struct dirty_log *log, region_t region)
{
	struct region_state *rs, *tmp_rs, *rs_new;
	struct log_c *lc = (struct log_c *) log->context;

	spin_lock(&lc->state_lock);

	/*
	 * An item on the clear_waiting list should have been flushed
	 * before getting this mark_region call.
	 */
	list_for_each_entry_safe(rs, tmp_rs, &lc->clear_waiting, rs_list)
		if (region == rs->rs_region) {
			DMERR("Region being marked found on clear_waiting list (%Lu/%s)",
			      region, lc->uuid + (strlen(lc->uuid) - 8));
			BUG();
		}

	/*
	 * We should never get two marks before a flush, unless the
	 * region is not in-sync.  One valid scenario would be:
	 *  0) region not in-sync
	 *  1) rh_inc (mark region)
	 *  2) rh_update_states
	 *  3) rh_dec (dec pending and put on clean_region list)
	 *  4) do_writes -> rh_inc (second mark)
	 */
	list_for_each_entry_safe(rs, tmp_rs, &lc->mark_waiting, rs_list)
		if (region == rs->rs_region)
			goto out;

	/*
	 * It is possible for the following in the mirror code:
	 *  0) Mark is already logged for a region
	 *  1) rh_update_states (were the clear_region happens)
	 *  2) rh_dec, sets region state to RH_CLEAN (asynchronous)
	 *  3) do_writes, trys to mark region
	 *
	 * This can lead to this next case being valid.
	 */
	list_for_each_entry_safe(rs, tmp_rs, &lc->mark_logged, rs_list)
		if (region == rs->rs_region)
			goto out;

	spin_unlock(&lc->state_lock);

	rs_new = mempool_alloc(region_state_pool, GFP_NOFS);
	BUG_ON(!rs_new);
	memset(rs_new, 0, sizeof(struct region_state));

	spin_lock(&lc->state_lock);
	rs_new->rs_mark_logged = 1;
	rs_new->rs_region = region;
	INIT_LIST_HEAD(&rs_new->rs_list);
	list_add(&rs_new->rs_list, &lc->mark_waiting);
out:
	spin_unlock(&lc->state_lock);

	return;
}

static void cluster_clear_region(struct dirty_log *log, region_t region)
{
	struct log_c *lc = (struct log_c *) log->context;
	struct region_state *rs, *tmp_rs, *rs_new;

	spin_lock(&lc->state_lock);

	/*
	 * The nominal case is to find the region in the marked list
	 */
	list_for_each_entry_safe(rs, tmp_rs, &lc->mark_logged, rs_list){
		if(region == rs->rs_region){
			list_del_init(&rs->rs_list);
			list_add(&rs->rs_list, &lc->clear_waiting);
			goto out;
		}
	}

	/*
	 * It is possible, but unlikely to get to this case. It requires
	 * the following to happen:
	 * 1) mark the region for writing
	 * 2) clear the region
	 * 3) clear doesn't get flushed because of bug 235040
	 * 4) suspend due to server relocation
	 * 5) on-disk log says we need to recover (because it hasn't been cleared)
	 * 6) we recover the region
	 * 7) clearing the region after recovery causes us to get here
	 *
	 * Bug 235040 cleared.  This should no longer happen.
	 */
	list_for_each_entry_safe(rs, tmp_rs, &lc->clear_waiting, rs_list){
		if(region == rs->rs_region){
			DMERR("Double clear on region (%Lu/%s)",
			      region, lc->uuid + (strlen(lc->uuid) - 8));
			BUG();
		}
	}

	list_for_each_entry_safe(rs, tmp_rs, &lc->mark_waiting, rs_list){
		if(region == rs->rs_region){
			DMERR("Clear pre-empting mark (%Lu/%s)",
			       region, lc->uuid + (strlen(lc->uuid) - 8));
			BUG();
		}
	}
	
	/* We can get here because we may be doing resync_work, and therefore,**
	** clearing without ever marking..................................... */

	/* Don't need to spin_unlock, because allocation is non-blocking */
	rs_new = mempool_alloc(region_state_pool, GFP_ATOMIC);
	if (!rs_new) {
		DMERR("Failed to allocate space for clear region request: %Lu",
		      region);
		BUG();
	}
	memset(rs_new, 0, sizeof(struct region_state));

	rs_new->rs_region = region;
	INIT_LIST_HEAD(&rs_new->rs_list);
	list_add(&rs_new->rs_list, &lc->clear_waiting);

out:
	spin_unlock(&lc->state_lock);

	return;
}

static int cluster_get_resync_work(struct dirty_log *log, region_t *region)
{
	int rtn;
	struct log_c *lc = (struct log_c *) log->context;

	rtn = consult_server(lc, 0, LRT_GET_RESYNC_WORK, region);

	if (*region > lc->region_count) {
		DMWARN("Error while getting resync work: bad region");
		rtn = 0;
	}

	if (rtn)
		DMDEBUG("Received recovery work from %u: %Lu/%s",
			lc->server_id, *region, lc->uuid + (strlen(lc->uuid) - 8));

	/*
	 * Check for bug 235039
	 * Note the changes in cluser_clear_region
	 */
	if (rtn == 1) {
		struct region_state *rs, *tmp_rs;
		list_for_each_entry_safe(rs, tmp_rs, &lc->clear_waiting, rs_list) {
			if (*region == rs->rs_region) {
				DMERR("WARNING: Bug 235039/235040 detected!");
				BUG();
			}
		}
	}

	return rtn;
}

static void cluster_complete_resync_work(struct dirty_log *log,
					 region_t region, int success)
{
	int i;
	region_t success_tmp = success;
	struct log_c *lc = (struct log_c *) log->context;

	if (success)
		DMDEBUG("Client finishing recovery: %Lu/%s",
			region, lc->uuid + (strlen(lc->uuid) - 8));
	else
		DMDEBUG("Notifying server(%u) of sync change: %Lu/%s",
			lc->server_id, region,
			lc->uuid + (strlen(lc->uuid) - 8));
	for (i = 0; i < 5; i++) {
		if (!consult_server(lc, region,
				    LRT_COMPLETE_RESYNC_WORK, &success_tmp))
			break;
		success_tmp = success;
		DMWARN("Unable to notify server of sync state change");
	}
	return;
}

static region_t cluster_get_sync_count(struct dirty_log *log)
{
	int i;
	region_t rtn;
	struct log_c *lc = (struct log_c *) log->context;

	if (atomic_read(&lc->suspended)) {
		return (atomic_read(&lc->in_sync) == 1) ? lc->region_count : 0;
	}

	/* Try to get sync count up to five times */
	for (i = 0; i < 5 && consult_server(lc, 0, LRT_GET_SYNC_COUNT, &rtn); i++);
	if(i >= 5){
		return 0;
	}

	if(rtn > lc->region_count){
		DMERR("sync_count ("
		      SECTOR_FORMAT
		      ") > region_count ("
		      SECTOR_FORMAT
		      ") - (%s)!",
		      rtn, lc->region_count, lc->uuid + (strlen(lc->uuid) - 8));
	}

	if(rtn >= lc->region_count){
		atomic_set(&lc->in_sync, 1);
	} else if(unlikely(atomic_read(&lc->in_sync) < 0)){
		atomic_set(&lc->in_sync, 0);
	}

	return rtn;
}

static int cluster_status(struct dirty_log *log, status_type_t status,
			  char *result, unsigned int maxlen)
{
	int sz = 0;
	int arg_count=2;
	struct log_c *lc = (struct log_c *) log->context;

	switch(status){
	case STATUSTYPE_INFO:
		if(lc->sync != DEFAULTSYNC)
			arg_count++;

		if (lc->log_dev)
			DMEMIT("3 %s %s %c",
			       log->type->name,                  /* NAME */
			       lc->log_dev->name,                /* THE LOG DEVICE */
			       (lc->log_dev_failed)? 'D' : 'A'); /* LOG DEVICE LIVENESS */
		else
			DMEMIT("1 %s", log->type->name);

                break;

        case STATUSTYPE_TABLE:
		DMDEBUG("LOG INFO:");
		DMDEBUG("  uuid: %s", lc->uuid);
		DMDEBUG("  uuid_ref    : %d", lc->uuid_ref);
		DMDEBUG("  log type    : %s", (lc->log_dev) ? "disk" : "core");
		DMDEBUG(" ?region_count: %Lu", lc->region_count);
		DMDEBUG(" ?sync_count  : %Lu", lc->sync_count);
		DMDEBUG(" ?sync_search : %d", lc->sync_search);
		DMDEBUG("  in_sync     : %s", (atomic_read(&lc->in_sync)) ? "YES" : "NO");
		DMDEBUG("  suspended   : %s", (atomic_read(&lc->suspended)) ? "YES" : "NO");
		DMDEBUG("  recovery_halted : %s", (lc->recovery_halted) ? "YES" : "NO");
		DMDEBUG("  server_id   : %u", lc->server_id);
		DMDEBUG("  server_valid: %s",
			((lc->server_id != 0xDEAD) && lc->server_valid) ? "YES" : "NO");
		{
			struct log_c *tmp_lc;

			down(&log_list_lock);
			list_for_each_entry(tmp_lc, &log_list_head, log_list){
				if(!strncmp(tmp_lc->uuid, lc->uuid, MAX_NAME_LEN) &&
				   (tmp_lc->uuid_ref != lc->uuid_ref)){
					DMDEBUG("LOG INFO [COPY FOUND]:");
					DMDEBUG("  uuid: %s", tmp_lc->uuid);
					DMDEBUG("  uuid_ref    : %d", tmp_lc->uuid_ref);
					DMDEBUG(" ?region_count: %Lu", tmp_lc->region_count);
					DMDEBUG(" ?sync_count  : %Lu", tmp_lc->sync_count);
					DMDEBUG(" ?sync_search : %d", tmp_lc->sync_search);
					DMDEBUG("  in_sync     : %s",
						(atomic_read(&tmp_lc->in_sync)) ? "YES" : "NO");
					DMDEBUG("  suspended   : %s",
						(atomic_read(&tmp_lc->suspended)) ? "YES" : "NO");
					DMDEBUG("  recovery_halted : %s",
						(tmp_lc->recovery_halted) ? "YES" : "NO");
					DMDEBUG("  server_id   : %u", tmp_lc->server_id);
					DMDEBUG("  server_valid: %s",
						((tmp_lc->server_id != 0xDEAD) &&
						 tmp_lc->server_valid) ? "YES" : "NO");
				}
			}
			up(&log_list_lock);
		}
		if(lc->sync != DEFAULTSYNC)
			arg_count++;

		if (lc->failure_response == FR_BLOCK)
			arg_count++;

		if (lc->log_dev) {
			arg_count++;

			DMEMIT("%s %u %s " SECTOR_FORMAT " %s ",
			       log->type->name,                 /* NAME */
			       arg_count,                       /* # OF ARGS */
			       lc->log_dev->name,               /* THE LOG DEVICE */
			       lc->region_size,                 /* REGION SIZE */
			       lc->uuid);                       /* UUID */
		} else {
			DMEMIT("%s %u " SECTOR_FORMAT " %s ",
			       log->type->name,                 /* NAME */
			       arg_count,                       /* # OF ARGS */
			       lc->region_size,                 /* REGION SIZE */
			       lc->uuid);                       /* UUID */
		}
		if (lc->sync != DEFAULTSYNC)
			DMEMIT("%ssync ", lc->sync == NOSYNC ? "no" : "");

		if (lc->failure_response == FR_BLOCK)
			DMEMIT("block_on_error ");
        }

	return sz;
}

static int cluster_get_failure_response(struct dirty_log *log)
{
	struct log_c *lc = (struct log_c *) log->context;
	return lc->failure_response;
}

static int clog_stop(void *data){
	struct log_c *lc;

	atomic_set(&suspend_client, 1);

	down(&log_list_lock);
	list_for_each_entry(lc, &log_list_head, log_list) {
		atomic_set(&lc->in_sync, 0);
	}
	up(&log_list_lock);

	return 0;
}

static int clog_start(void *data, uint32_t *nodeids, int count, int event_id, int type){
	int i;
	uint32_t server;
	struct log_c *lc;
	struct kcl_cluster_node node;

	if(global_nodeids){
		kfree(global_nodeids);
	}
	global_nodeids = nodeids;
	global_count = count;

	for (i = 0; kcl_get_node_by_nodeid(0, &node); i++) {
		if (i > 10)
			BUG();
		else
			DMERR("Bad call to kcl_get_node_by_nodeid");
	}
	my_id = node.node_id;

	/* Wait for any outstanding starts to complete */
	suspend_on(&event_queue, atomic_read(&restart_event_type));

	restart_event_id = event_id;
	atomic_set(&restart_event_type, type);

	switch(type){
	case SERVICE_NODE_LEAVE:
	case SERVICE_NODE_FAILED:
		down(&log_list_lock);
		list_for_each_entry(lc, &log_list_head, log_list){
			for(i=0, server = 0xDEAD; i < count; i++){
				if(lc->server_id == nodeids[i]){
					server = nodeids[i];
				}
			}
			/* ATTENTION -- need locking around this ? */
			lc->server_id = server;
		}
		up(&log_list_lock);

		break;
	case SERVICE_NODE_JOIN:
		break;
	default:
		DMERR("Invalid service event type received");
		BUG();
		break;
	}

	return 0;
}

static void clog_finish(void *data, int event_id){
	atomic_set(&suspend_client, 0);
	wake_up_all(&suspend_client_queue);
}

static struct kcl_service_ops clog_ops = {
	.stop = clog_stop,
	.start = clog_start,
	.finish = clog_finish,
};

static int mirror_set_count = 0; /* used to prevent multiple cluster [dis]connects */
static DECLARE_MUTEX(cmirror_register_lock);

static int cluster_connect(void)
{
	int r = 0;

	down(&cmirror_register_lock);

	if (mirror_set_count++)
		goto out;

	r = kcl_register_service("clustered_log", 13, SERVICE_LEVEL_GDLM, &clog_ops,
				 1, NULL, &local_id);
	if (r) {
		DMWARN("Couldn't register clustered_log service.  Reason: %d", r);
		goto out;
	}

	r = start_server();
	if(r){
		DMWARN("Unable to start clustered log server daemon");
		kcl_unregister_service(local_id);
		goto out;
	}

	r = kcl_join_service(local_id);

	if(r){
		DMWARN("couldn't join service group");
		stop_server();
		kcl_unregister_service(local_id);
	}

out:
	up(&cmirror_register_lock);
	return r;
}

static int cluster_disconnect(void)
{
	down(&cmirror_register_lock);

	if (--mirror_set_count) {
		up(&cmirror_register_lock);
		return 0;
	}

	/* By setting 'shutting_down', the server will not be suspended **
	** when a stop is received */
	shutting_down = 1;
	kcl_leave_service(local_id);
	stop_server();
	kcl_unregister_service(local_id);

	up(&cmirror_register_lock);
	return 0;
}

static struct dirty_log_type _clustered_core_type = {
	.name = "clustered_core",
	.module = THIS_MODULE,
	.ctr = cluster_core_ctr,
	.dtr = cluster_dtr,
	.presuspend = cluster_presuspend,
	.postsuspend = cluster_postsuspend,
	.resume = cluster_resume,
	.get_region_size = cluster_get_region_size,
	.is_clean = cluster_is_clean,
	.is_remote_recovering = cluster_is_remote_recovering,
	.in_sync = cluster_in_sync,
	.flush = cluster_flush,
	.mark_region = cluster_mark_region,
	.clear_region = cluster_clear_region,
	.get_resync_work = cluster_get_resync_work,
	.complete_resync_work = cluster_complete_resync_work,
	.get_sync_count = cluster_get_sync_count,
	.status = cluster_status,
	.get_failure_response = cluster_get_failure_response,
};

static struct dirty_log_type _clustered_disk_type = {
	.name = "clustered_disk",
	.module = THIS_MODULE,
	.ctr = cluster_disk_ctr,
	.dtr = cluster_dtr,
	.presuspend = cluster_presuspend,
	.postsuspend = cluster_postsuspend,
	.resume = cluster_resume,
	.get_region_size = cluster_get_region_size,
	.is_clean = cluster_is_clean,
	.is_remote_recovering = cluster_is_remote_recovering,
	.in_sync = cluster_in_sync,
	.flush = cluster_flush,
	.mark_region = cluster_mark_region,
	.clear_region = cluster_clear_region,
	.get_resync_work = cluster_get_resync_work,
	.complete_resync_work = cluster_complete_resync_work,
	.get_sync_count = cluster_get_sync_count,
	.status = cluster_status,
	.get_failure_response = cluster_get_failure_response,
};

#define CMIRROR_RELEASE_NAME "0.2.0"
static int __init cluster_dirty_log_init(void)
{
	int r = 0;

        DMINFO("dm-cmirror %s (built %s %s) installed",
               CMIRROR_RELEASE_NAME, __DATE__, __TIME__);

	region_state_pool = mempool_create(500, region_state_alloc,
					   region_state_free, NULL);
	if(!region_state_pool){
		DMWARN("couldn't create region state pool");
		return -ENOMEM;
	}

	init_waitqueue_head(&suspend_client_queue);
	init_waitqueue_head(&event_queue);

	r = dm_register_dirty_log_type(&_clustered_core_type);
	if (r) {
		DMWARN("couldn't register clustered_core dirty log type");
		mempool_destroy(region_state_pool);
		return r;
	}

	r = dm_register_dirty_log_type(&_clustered_disk_type);
	if (r) {
		DMWARN("couldn't register clustered_disk dirty log type");
		dm_unregister_dirty_log_type(&_clustered_core_type);
		mempool_destroy(region_state_pool);
		return r;
	}

	return r;

}

static void __exit cluster_dirty_log_exit(void)
{
	if(!list_empty(&log_list_head)){
		DMERR("attempt to remove module, but dirty logs are still in place!");
		DMERR("this is a fatal error");
		BUG();
	}
	dm_unregister_dirty_log_type(&_clustered_core_type);
	dm_unregister_dirty_log_type(&_clustered_disk_type);
	mempool_destroy(region_state_pool);
        DMINFO("dm-cmirror %s (built %s %s) removed",
               CMIRROR_RELEASE_NAME, __DATE__, __TIME__);
}

module_init(cluster_dirty_log_init);
module_exit(cluster_dirty_log_exit);

MODULE_DESCRIPTION(DM_NAME " cluster capable mirror logs (clustered mirroring)");
MODULE_AUTHOR("Jonathan Brassow");
MODULE_LICENSE("GPL");

/*
 * Overrides for Emacs so that we follow Linus's tabbing style.
 * Emacs will notice this stuff at the end of the file and automatically
 * adjust the settings for this buffer only.  This must remain at the end
 * of the file.
 * ---------------------------------------------------------------------------
 * Local variables:
 * c-file-style: "linux"
 * End:
 */
