/*
 * Copyright (C) 2005 Red Hat, Inc.
 *
 * This file is released under the LGPL.
 */

#include <linux/init.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/vmalloc.h>
#include <linux/list.h>
#include <linux/net.h>
#include <linux/in.h>
#include <linux/socket.h>
#include <linux/signal.h>
#include <linux/bio.h>
#include <linux/blkdev.h>
#include <linux/device-mapper.h>
#include <cluster/service.h>
#include <cluster/cnxman.h>
#include <cluster/cnxman-socket.h>

#include "dm-log.h"
#include "dm-cmirror-common.h"
#include "dm-cmirror-xfr.h"
#include "dm-cmirror-cman.h"

#define RU_READ    0
#define RU_WRITE   1
#define RU_RECOVER 2
#define BYTE_SHIFT 3

struct region_user {
	struct list_head ru_list;
	int32_t  ru_rw;
	uint32_t ru_nodeid;
	region_t ru_region;
};

static mempool_t *region_user_pool;

static atomic_t server_run;
static struct completion server_completion;

static atomic_t _do_requests;

static int debug_disk_write = 0;
extern struct semaphore log_list_lock;
extern struct list_head log_list_head;

static void *region_user_alloc(int gfp_mask, void *pool_data){
	return kmalloc(sizeof(struct region_user), gfp_mask);
}

static void region_user_free(void *element, void *pool_data){
	kfree(element);
}

/*
 * The touched member needs to be updated every time we access
 * one of the bitsets.
 */
static inline int log_test_bit(uint32_t *bs, unsigned bit)
{
	return ext2_test_bit(bit, (unsigned long *) bs) ? 1 : 0;
}

static inline void log_set_bit(struct log_c *l,
			       uint32_t *bs, unsigned bit)
{
	ext2_set_bit(bit, (unsigned long *) bs);
	l->touched = 1;
}

static inline void log_clear_bit(struct log_c *l,
				 uint32_t *bs, unsigned bit)
{
	ext2_clear_bit(bit, (unsigned long *) bs);
	l->touched = 1;
}

/*----------------------------------------------------------------
 * Header IO
 *--------------------------------------------------------------*/
static void header_to_disk(struct log_header *core, struct log_header *disk)
{
	disk->magic = cpu_to_le32(core->magic);
	disk->version = cpu_to_le32(core->version);
	disk->nr_regions = cpu_to_le64(core->nr_regions);
	memcpy(disk->uuid, core->uuid, MAX_NAME_LEN);
}

static void header_from_disk(struct log_header *core, struct log_header *disk)
{
	core->magic = le32_to_cpu(disk->magic);
	core->version = le32_to_cpu(disk->version);
	core->nr_regions = le64_to_cpu(disk->nr_regions);
	memcpy(core->uuid, disk->uuid, MAX_NAME_LEN);
}

static void do_io_callback(unsigned long error, void *context)
{
	struct log_c *lc = context;

	lc->log_error = (error) ? -EIO : 0;
	complete(&lc->complete);
}

/*
 * do_io
 * @log
 * @rw: read/write
 * @header: header (if not, then bits)
 *
 * Returns: 0 on success, -EXXX on error
 */
static int do_io(struct log_c *log, int rw, int header)
{
	int r;

	r = dm_io_async_vm(1,
			   (header) ? &log->header_location : &log->bits_location,
			   rw,
			   (header) ? (void *)(log->disk_header) : (void *)(log->clean_bits),
			   do_io_callback, log);
	if (r) {
		DMERR("Error while submitting log I/O: %d", r);
		return r;
	}

	r = wait_for_completion_timeout(&log->complete, 5 * HZ);
	if (!r) {
		DMERR("[%s] Timed out waiting for log I/O: %s of %s",
		      SHORT_UUID(log->uuid), (rw == WRITE) ? "Write" : "Read",
		      (header) ? "header" : "bits");
		return -EDEADLK;
	}
	return log->log_error;
}

int read_header(struct log_c *log)
{
	int r;

	if (!log->log_dev)
		return 0;

	if (atomic_read(&log->suspended))
		return -EDEADLK;

	r = do_io(log, READ, 1);
	if (unlikely(r))
		return r;

	header_from_disk(&log->header, log->disk_header);

	/* New log required? */
	if (log->sync != DEFAULTSYNC || log->header.magic != MIRROR_MAGIC) {
		log->header.magic = MIRROR_MAGIC;
		log->header.version = MIRROR_DISK_VERSION;
		log->header.nr_regions = 0;
		memset(log->header.uuid, 0, MAX_NAME_LEN);
	}

	if (log->header.version != MIRROR_DISK_VERSION) {
		DMWARN("incompatible disk log version");
		return -EINVAL;
	}

	return 0;
}

int write_header(struct log_c *log)
{
	if (!log->log_dev)
		return 0;

	if (atomic_read(&log->suspended))
		return -EDEADLK;

	header_to_disk(&log->header, log->disk_header);

	return do_io(log, WRITE, 1);
}

/*----------------------------------------------------------------
 * Bits IO
 *--------------------------------------------------------------*/
static inline void zeros_to_core(uint32_t *core, unsigned count)
{
	memset(core, 0, sizeof(uint32_t)*count);
}

static inline void bits_to_core(uint32_t *core, uint32_t *disk, unsigned count)
{
	unsigned i;

	for (i = 0; i < count; i++)
		core[i] = le32_to_cpu(disk[i]);
}

static inline void zeros_to_disk(uint32_t *disk, unsigned count)
{
	memset(disk, 0, sizeof(uint32_t)*count);
}

static inline void bits_to_disk(uint32_t *core, uint32_t *disk, unsigned count)
{
	unsigned i;

	/* copy across the clean/dirty bitset */
	for (i = 0; i < count; i++)
		disk[i] = cpu_to_le32(core[i]);
}

static int read_bits(struct log_c *log)
{
	int r;

	if (!log->log_dev)
		return 0;

	if (atomic_read(&log->suspended))
		return -EDEADLK;

	r = do_io(log, READ, 0);

	if (unlikely(r))
		return r;

	return 0;
}

static int write_bits(struct log_c *log)
{
	if (!log->log_dev)
		return 0;

	if (atomic_read(&log->suspended))
		return -EDEADLK;

	return do_io(log, WRITE, 0);
}

static int count_bits32(uint32_t *addr, unsigned size)
{
	int count = 0, i;

	for (i = 0; i < size; i++) {
		count += hweight32(*(addr+i));
	}
	return count;
}

static int print_zero_bits(unsigned char *str, int offset, int bit_count){
	int i,j;
	int count=0;
	int len = bit_count/8 + ((bit_count%8)?1:0);
	int region = offset;
	int range_count=0;

	for(i = 0; i < len; i++){
		if(str[i] == 0x0){
			region+=(bit_count < 8)? bit_count: 8;
			range_count+= (bit_count < 8)? bit_count: 8;
			count+=(bit_count < 8)? bit_count: 8;

			bit_count -= (bit_count < 8)? bit_count: 8;
			continue;
		} else if(str[i] == 0xFF){
			if(range_count==1){
				DMDEBUG("  %d", region - 1);
			} else if(range_count){
				DMDEBUG("  %d - %d", region-range_count, region-1);
			}
			range_count = 0;
			region+=(bit_count < 8)? bit_count: 8;      

			bit_count -= (bit_count < 8)? bit_count: 8;
			continue;
		}
		for(j=0; j<8; j++){
			if(!bit_count--){
				break;
			}
			if(!(str[i] & 1<<j)){
				range_count++;
				region++;
				count++;
			} else {
				if(range_count==1){
					DMDEBUG("  %d", region - 1);
				} else if(range_count){
					DMDEBUG("  %d - %d", region-range_count, region-1);
				}
				range_count = 0;
				region++;
			}
		}
	}

	if(range_count==1){
		DMDEBUG("  %d", region - 1);
	} else if(range_count){
		DMDEBUG("  %d - %d", region-range_count, region);
	}
	return count;
}

static int disk_resume(struct log_c *lc)
{
	int r;
	int good_count=0, bad_count=0;
	unsigned i;
	size_t size = lc->bitset_uint32_count * sizeof(uint32_t);
	struct region_user *tmp_ru, *ru;
	unsigned char live_nodes[16]; /* Attention -- max of 128 nodes... */

	DMDEBUG("Disk Resume::  %s (%s)", lc->uuid + (strlen(lc->uuid) - 8),
		atomic_read(&lc->suspended) ? "suspended" : "active");

	debug_disk_write = 1;
	memset(live_nodes, 0, sizeof(live_nodes));
	for(i = 0; i < global_count; i++){
		live_nodes[global_nodeids[i]/8] |= 1 << (global_nodeids[i]%8);
	}

	/* read the disk header */
	i = 1;
	if (!lc->log_dev_failed &&
	    ((r = read_header(lc)) || (i = 0) || (r = read_bits(lc)))) {
		if (r == -EINVAL || r == -EDEADLK)
			return r;

		DMWARN("Read %s failed on mirror log device, %s",
		       i ? "header" : "bits", lc->log_dev->name);
		lc->log_dev_failed = 1;
		lc->header.nr_regions = 0;
	}

	/* set or clear any new bits -- device has grown */
	if (lc->sync == NOSYNC) {
		DMDEBUG("  NOSYNC            :: set");
		for (i = lc->header.nr_regions; i < lc->region_count; i++)
			/* FIXME: amazingly inefficient */
			log_set_bit(lc, lc->clean_bits, i);
	} else {
		DMDEBUG("  NOSYNC            :: unset");
		for (i = lc->header.nr_regions; i < lc->region_count; i++)
			/* FIXME: amazingly inefficient */
			log_clear_bit(lc, lc->clean_bits, i);
	}

	/* clear any old/unused bits -- device has shrunk */
	for(i = lc->region_count; i % (sizeof(*lc->clean_bits) << BYTE_SHIFT); i++)
		log_clear_bit(lc, lc->clean_bits, i);

	/* copy clean across to sync */
	memcpy(lc->sync_bits, lc->clean_bits, size);

	/* must go through the list twice.  The dead node could have been using **
	** the same region as other nodes and we want any region that was in    **
	** use by the dead node to be marked _not_ in-sync..................... */
	lc->recovering_region = (uint64_t)-1;
	list_for_each_entry(ru, &lc->region_users, ru_list){
		if (live_nodes[ru->ru_nodeid/8] & 1 << (ru->ru_nodeid%8)) {
			good_count++;
			if (ru->ru_rw == RU_WRITE) {
				log_set_bit(lc, lc->sync_bits, ru->ru_region);
			} else if (ru->ru_rw == RU_RECOVER) {
				log_clear_bit(lc, lc->sync_bits, ru->ru_region);
				lc->recovering_region = ru->ru_region;
			}
		}
	}

	list_for_each_entry_safe(ru, tmp_ru, &lc->region_users, ru_list){
		if(!(live_nodes[ru->ru_nodeid/8] & 1 << (ru->ru_nodeid%8))){
			bad_count++;
			log_clear_bit(lc, lc->sync_bits, ru->ru_region);
			if (ru->ru_rw == RU_RECOVER) {
				DMINFO("Failed node was recovering region %Lu - cleared",
				       ru->ru_region);
				lc->recovering_region = (uint64_t)-1;
			}
			list_del(&ru->ru_list);
			mempool_free(ru, region_user_pool);
		}
	}

	DMDEBUG("  Live nodes        :: %d", global_count);
	DMDEBUG("  In-Use Regions    :: %d", good_count+bad_count);
	DMDEBUG("  Good IUR's        :: %d", good_count);
	DMDEBUG("  Bad IUR's         :: %d", bad_count);

	lc->sync_count = count_bits32(lc->sync_bits, lc->bitset_uint32_count);
	lc->sync_search = 0;

	DMDEBUG("  in_sync           :: %d", atomic_read(&lc->in_sync));
	DMDEBUG("  Sync count        :: %Lu", lc->sync_count);
	DMDEBUG("  Disk Region count :: %Lu", lc->header.nr_regions);
	DMDEBUG("  Region count      :: %Lu", lc->region_count);

	if (lc->header.nr_regions != lc->region_count) {
		DMDEBUG("  NOTE:  Mapping has changed.");
	}
/* Take this out for now.
	if(list_empty(&lc->region_users) && (lc->sync_count != lc->header.nr_regions)){
		struct region_user *new;
		
		for(sync_search = 0; sync_search < lc->header.nr_regions;){
			region = find_next_zero_bit((unsigned long *)lc->clean_bits,
						    lc->header.nr_regions,
						    sync_search);
			sync_search = region+1;
			if(region < lc->header.nr_regions){
				for(i=0; i < global_count; i++){
					new = kmalloc(sizeof(struct region_user),
						      GFP_KERNEL);
					if(!new){
						DMERR("Unable to allocate space to track region users.");
						BUG();
					}
					new->ru_nodeid = global_nodeids[i];
					new->ru_region = region;
					DMINFO("Adding %u/%Lu",
					       new->ru_nodeid, new->ru_region);
					list_add(&new->ru_list, &lc->region_users);
				}
			}
		}
	}			

*/
	DMDEBUG("Marked regions::");
	i = print_zero_bits((unsigned char *)lc->clean_bits, 0, lc->region_count);
	DMDEBUG("  Total = %d", i);

	DMDEBUG("Out-of-sync regions::");
	i = print_zero_bits((unsigned char *)lc->sync_bits, 0, lc->region_count);
	DMDEBUG("  Total = %d", i);

	/* set the correct number of regions in the header */
	lc->header.nr_regions = lc->region_count;

	i = 1;
	if ((r = write_bits(lc)) || (i = 0) || (r = write_header(lc))) {
		if (r != -EDEADLK) {
			DMWARN("Write %s failed on mirror log device, %s.",
			       i ? "bits" : "header", lc->log_dev->name);
			lc->log_dev_failed = 1;
		}
	} else 
		lc->log_dev_failed = 0;

	return r;
}


struct region_user *find_ru(struct log_c *lc, uint32_t who, region_t region){
	struct region_user *ru;
	list_for_each_entry(ru, &lc->region_users, ru_list){
		if((who == ru->ru_nodeid) && (region == ru->ru_region)){
			return ru;
		}
	}
	return NULL;
}

struct region_user *find_ru_by_region(struct log_c *lc, region_t region){
	struct region_user *ru;
	list_for_each_entry(ru, &lc->region_users, ru_list){
		if(region == ru->ru_region){
			return ru;
		}
	}
	return NULL;
}


static int server_is_clean(struct log_c *lc, struct log_request *lr)
{
	lr->u.lr_int_rtn = log_test_bit(lc->clean_bits, lr->u.lr_region);

	return 0;
}

static int server_is_remote_recovering(struct log_c *lc, struct log_request *lr)
{
	lr->u.lr_int_rtn = 0;

	if ((lc->sync_search > lc->region_count) && !lc->sync_pass)
		return 0;

	if (lc->recovery_halted &&
	    (lc->recovering_region != lr->u.lr_region)) {
		DMDEBUG("Recovery halted, allowing client: %Lu/%s",
			lr->u.lr_region,
			lc->uuid + (strlen(lc->uuid) - 8));
		return 0;
	}

	/*
	 * If the region hasn't been recovered yet,
	 * we need to block the write
	 */
	if (!log_test_bit(lc->sync_bits, lr->u.lr_region) ||
	    (lc->recovering_region == lr->u.lr_region)) {
		lr->u.lr_int_rtn = 1;

		/* Try to make this region a priority */
		if ((lr->u.lr_region != lc->recovering_region) &&
		    ((lc->recovering_next == (uint64_t)-1) ||
		     (lc->recovering_next > lr->u.lr_region)))
			lc->recovering_next = lr->u.lr_region;		
		return 0;
	}

	return 0;
}

static int server_in_sync(struct log_c *lc, struct log_request *lr)
{
	if (lr->u.lr_region > lc->region_count) {
		lr->u.lr_int_rtn = 0;
		return -EINVAL;
	}

	if(likely(log_test_bit(lc->sync_bits, lr->u.lr_region)))
		/* in-sync */
		lr->u.lr_int_rtn = 1;
	else
		lr->u.lr_int_rtn = 0;

	return 0;
}


static int server_mark_region(struct log_c *lc, struct log_request *lr, uint32_t who)
{
	struct region_user *ru, *new;

	new = mempool_alloc(region_user_pool, GFP_NOFS);
	if(!new){
		return -ENOMEM;
	}

	new->ru_nodeid = who;
	new->ru_region = lr->u.lr_region;
	new->ru_rw = RU_WRITE;

	if (find_ru(lc, who, lr->u.lr_region)) {
		DMWARN("Attempt to mark a already marked region (%u,"
		       SECTOR_FORMAT
		       "/%s)",
		       who, lr->u.lr_region, lc->uuid + (strlen(lc->uuid) - 8));
		mempool_free(new, region_user_pool);
	} else if (!(ru = find_ru_by_region(lc, lr->u.lr_region))) {
		log_clear_bit(lc, lc->clean_bits, lr->u.lr_region);
		list_add(&new->ru_list, &lc->region_users);
	} else if (ru->ru_rw == RU_RECOVER) {
		/*
		 * A mark that happens to a region in recovery
		 * means certain corruption.
		 */
		DMERR("Mark attempted to recovering region by %u: %Lu/%s",
		      who, lr->u.lr_region,
		      lc->uuid + (strlen(lc->uuid) - 8));
		DMERR("  region_user { rw=%s, nodeid=%u, region=%Lu }",
		      (ru->ru_rw == RU_WRITE) ? "RU_WRITE":
		      (ru->ru_rw == RU_RECOVER) ? "RU_RECOVER":
		      (ru->ru_rw == RU_READ) ? "RU_READ" : "UNKOWN",
		      ru->ru_nodeid, ru->ru_region);
		DMERR("  lc->recovering_region = %Lu", lc->recovering_region);
		DMERR("  lc->sync_count = %Lu", lc->sync_count);
		DMERR("  lc->in_sync = %d", atomic_read(&lc->in_sync));
		DMERR("  lc->sync_pass = %d", lc->sync_pass);
		DMERR("  lc->sync_search = %d", lc->sync_search);
		DMERR("  lc->recovery_halted = %d -> 1", lc->recovery_halted);

		lc->recovery_halted = 1;
		mempool_free(new, region_user_pool);
		return -EAGAIN;
	} else {
		list_add(&new->ru_list, &ru->ru_list);
	}

	return 0;
}


static int server_clear_region(struct log_c *lc, struct log_request *lr, uint32_t who)
{
	int check_bug = 0;
	struct region_user *ru;

	ru = find_ru(lc, who, lr->u.lr_region);
	if(!ru){
		DMDEBUG("Request to remove unrecorded region user (%u/%Lu)",
		       who, lr->u.lr_region);
		return -EINVAL;
	} else {
		if (lc->recovering_region == lr->u.lr_region) {
			lc->recovering_region = (uint64_t)-1;
			check_bug = 1;
		}
		list_del(&ru->ru_list);
		mempool_free(ru, region_user_pool);
	}

	if (!find_ru_by_region(lc, lr->u.lr_region)) {
		/* Only clear the region if it is also in sync */
		if (log_test_bit(lc->sync_bits, lr->u.lr_region))
			log_set_bit(lc, lc->clean_bits, lr->u.lr_region);
	} else if (check_bug) {
		DMERR("Multiple marks exist on a region being recovered: %Lu/%s",
		      lr->u.lr_region, lc->uuid + (strlen(lc->uuid) - 8));
		BUG();
	}
		
	return 0;
}


static int server_flush(struct log_c *lc, uint32_t who)
{
	int r = 0;
	struct region_user *ru;

	if (lc->recovering_region != (uint64_t)-1) {
		list_for_each_entry(ru, &lc->region_users, ru_list) {
			if ((ru->ru_region == lc->recovering_region) &&
			    (ru->ru_rw != RU_RECOVER)) {
				DMERR("Flush attempted to recovering region by %u: %Lu/%s",
				      who, lc->recovering_region,
				      lc->uuid + (strlen(lc->uuid) - 8));
				BUG();
			}
		}
	}

	r = write_bits(lc);
	if (!r) {
		lc->touched = 0;
		lc->log_dev_failed = 0;
	} else {
		lc->log_dev_failed = 1;
	}

	return (lc->log_dev_failed) ? -EIO : 0;
}


static int server_get_resync_work(struct log_c *lc, struct log_request *lr, uint32_t who)
{
	struct region_user *new, *test;
	region_t *region = &(lr->u.lr_region_rtn);
	region_t sync_search = lc->sync_search;

	lr->u.lr_int_rtn = 0; /* Default to no work */

	if (lc->recovery_halted) {
		DMDEBUG("Recovery halted due to error on %s",
			lc->uuid + (strlen(lc->uuid) - 8));
		return 0;
	}

	if (lc->recovering_region != (uint64_t)-1) {
		DMDEBUG("Someone is already recovering region %Lu/%s",
			lc->recovering_region, lc->uuid + (strlen(lc->uuid) - 8));
		return 0;
	}

	if (lc->sync_search >= lc->region_count) {
		/*
		 * FIXME: pvmove is not supported yet, but when it is,
		 * an audit of sync_count changes will need to be made
		 */
		if ((lc->sync_count < lc->region_count) && !lc->sync_pass) {
			lc->sync_search = 0;
			lc->sync_pass++;
		} else {
			return 0;
		}
	}

	if ((lc->recovering_next != (uint64_t)-1) &&
	    (!log_test_bit(lc->sync_bits, lc->recovering_next))) {
		*region = lc->recovering_next;
		DMDEBUG("Preempting normal recovery work for preferred region...");
	} else {
		*region = ext2_find_next_zero_bit((unsigned long *) lc->sync_bits,
						  lc->region_count,
						  lc->sync_search);
		sync_search = *region + 1;
	}

	if ((test = find_ru_by_region(lc, *region))) {
		/*
		 * We disallow writes to regions that have not yet been
		 * recovered via is_remote_recovering(), so this should
		 * not happen.
		 */
		DMERR("Recovery blocked by outstanding write on region %Lu/%s",
		      *region, lc->uuid + (strlen(lc->uuid) - 8));
		DMERR("  region_user { %s, %u, %Lu }",
		      (test->ru_rw == RU_WRITE) ? "RU_WRITE":
		      (test->ru_rw == RU_RECOVER) ? "RU_RECOVER":
		      (test->ru_rw == RU_READ) ? "RU_READ" : "UNKOWN",
		      test->ru_nodeid, test->ru_region);
		DMERR("  lc->recovering_region = %Lu", lc->recovering_region);
		DMERR("  lc->sync_count = %Lu", lc->sync_count);
		DMERR("  lc->in_sync = %d", atomic_read(&lc->in_sync));
		DMERR("  lc->sync_pass = %d", lc->sync_pass);
		DMERR("  lc->sync_search = %d", lc->sync_search);
		DMERR("  lc->recovery_halted = %d -> 1", lc->recovery_halted);

		lc->recovery_halted = 1;
		return 0;
	}

	if (*region >= lc->region_count)
		return 0;

	new = mempool_alloc(region_user_pool, GFP_NOFS);
	if (!new)
		return -ENOMEM;

	lc->sync_search = sync_search;
	lc->recovering_region = *region;
	lc->recovering_next = (uint64_t)-1;
	lr->u.lr_int_rtn = 1; /* Assigning work */
	new->ru_nodeid = who;
	new->ru_region = *region;
	new->ru_rw = RU_RECOVER;
	list_add(&new->ru_list, &lc->region_users);
	DMDEBUG("Assigning recovery work to %u: %Lu/%s",
		who, new->ru_region, lc->uuid + (strlen(lc->uuid) - 8));

	return 0;
}

static int server_complete_resync_work(struct log_c *lc, struct log_request *lr,
				       int success, uint32_t who){
	struct region_user *ru;

	if (lr->u.lr_region > lc->region_count) {
		return -EINVAL;
	}

	if (success) {
		if (lr->u.lr_region != lc->recovering_region) {
			DMERR("Told to clear recovery on wrong region %Lu/%Lu",
			      lr->u.lr_region, lc->recovering_region);
			return -EINVAL;
		}

		/* We could receive multiple identical request due to network failure */
		if (!log_test_bit(lc->sync_bits, lr->u.lr_region)) {
			log_set_bit(lc, lc->sync_bits, lr->u.lr_region);
			lc->sync_count++;
		}

		/*
		 * We will: 
		 * lc->recovering_region = (uint64_t)-1;
		 * in clear_region so we can do extra validation
		 */

		lc->sync_pass = 0;

		DMDEBUG("Resync work completed by %u: %Lu/%s",
			who, lr->u.lr_region, lc->uuid + (strlen(lc->uuid) - 8));
		return 0;
	}

	/*
	 * Recovery failed or mirror is being marked out-of-sync
	 *
	 * We need to stop dishing out recovery work.  If we don't
	 * writes happening to NOSYNC regions can't proceed and the
	 * mirror won't be able to suspend for reconfiguration - due
	 * to the return of is_remote_recovering().
	 *
	 * We can recieve multiple calls to mark out-of-sync
	 * if there were several writes to the same region that
	 * failed.  In this case, there will not be a record for
	 * the region.
	 */
	if (!lc->recovery_halted) {
		DMDEBUG("server_complete_resync_work - Setting recovery_halted = 1");
		lc->recovery_halted = 1;
	}

	ru = find_ru(lc, who, lr->u.lr_region);

	if ((lr->u.lr_region == lc->recovering_region) && !ru) {
		DMERR("Unable to locate record of recovery");
		BUG();
	}

	if (!ru) {
		DMDEBUG("Unable to find region to be marked out-of-sync: %Lu/%s/%u",
			lr->u.lr_region, lc->uuid + (strlen(lc->uuid) - 8), who);
		/*
		 * This is a valid case, when the following happens:
		 * 1) a region is recovering and has waiting writes
		 * 2) recovery fails and calls complete_resync_work (w/ failure)
		 * 2.1) RU is removed from our list
		 * 3) waiting writes are released
		 * 3.1) writes do not mark, because b/c region state != RH_CLEAN
		 * 4) write fails and calls complete_resync_work (w/ failure)
		 * 5) boom, we are here.
		 *
		 * Not a bug to be here
		 */
	} else 	if (ru->ru_rw == RU_RECOVER) {
		if (lr->u.lr_region != lc->recovering_region) {
			DMERR("Recovering region mismatch from node %u: (%Lu/%Lu)",
			      who, lr->u.lr_region, lc->recovering_region);
			BUG();
		}
		/*
		 * Clear the recovery
		 */
		DMDEBUG("Setting recovering region out-of-sync: %Lu/%s/%u",
			lr->u.lr_region, lc->uuid + (strlen(lc->uuid) - 8), who);
		lc->recovering_region = (uint64_t)-1;
		list_del(&ru->ru_list);
		mempool_free(ru, region_user_pool);
	}

	/* else if (ru->ru_rw == RU_WRITE)
	 * Mirror has place the region into RH_NOSYNC
	 * We will leave the record in place.  It is
	 * likely there are more requests coming to
	 * mark this region out-of-sync, due to the
	 * way mirror handles the situation.
	 *
	 DMDEBUG("Setting marked region out-of-sync: %Lu/%s",
                 lr->u.lr_region, lc->uuid + (strlen(lc->uuid) - 8));
	 */

	/* gone again: lc->sync_count--;*/
	log_clear_bit(lc, lc->sync_bits, lr->u.lr_region);

	return 0;
}

static int server_get_sync_count(struct log_c *lc, struct log_request *lr){
	if (lc->sync_count > lc->region_count) {
		DMERR("sync_count (" SECTOR_FORMAT ") > region_count (" SECTOR_FORMAT ") in %s!",
		      lc->sync_count, lc->region_count, lc->uuid + (strlen(lc->uuid) - 8));
		disk_resume(lc);
	}

	lr->u.lr_region_rtn = lc->sync_count;
	return 0;
}

static void report_duplicate_log(struct log_c *lc)
{
	DMERR("HEY!!! There are two matches for %s",
	      lc->uuid + (strlen(lc->uuid) - 8));
	list_for_each_entry(lc, &log_list_head, log_list) {
		DMERR("  %s", lc->uuid + (strlen(lc->uuid) - 8));
	}
}

static struct log_c *get_log_context(char *uuid, int uuid_ref){
	struct log_c *lc, *r = NULL;

	down(&log_list_lock);
	list_for_each_entry(lc, &log_list_head, log_list){
		if (!strncmp(lc->uuid, uuid, MAX_NAME_LEN) &&
		    (uuid_ref == lc->uuid_ref)) {
			if (r)
				report_duplicate_log(lc);
			else
				r = lc;
		}
	}
	up(&log_list_lock);

	return r;
}


static int process_election(struct log_request *lr, struct log_c *lc,
			    struct sockaddr_in *saddr){
	int i;
	uint32_t lowest, next;
	uint32_t node_count=global_count, *nodeids=global_nodeids;

	DMDEBUG("%s(%d): (%s)", RQ_STRING(lr->lr_type), lr->lr_type,
		(lc) ? lc->uuid + (strlen(lc->uuid) - 8) : "none");
	DMDEBUG("  starter     : %u", lr->u.lr_starter);
	DMDEBUG("  co-ordinator: %u", lr->u.lr_coordinator);
	DMDEBUG("  node_count  : %d", lr->u.lr_node_count);

	/* Record the starter's port number so we can get back to him */
	if((lr->u.lr_starter == my_id) && (!lr->u.lr_node_count)){
		lr->u.lr_starter_port = saddr->sin_port;
	}

	/* Find the next node id in the circle */
	for(lowest = my_id, next = my_id, i=0; i < node_count; i++){
		if(lowest > nodeids[i]){
			lowest = nodeids[i];
		}
		if(((next == my_id) || (next > nodeids[i])) &&
		   (nodeids[i] > my_id)){
			next = nodeids[i];
		}
	}

	/* Set address to point to next node in the circle */
	next = (next == my_id)? lowest: next;
	saddr->sin_port = CLUSTER_LOG_PORT;
	if(!(saddr->sin_addr.s_addr = nodeid_to_ipaddr(next))){
		return -1;
	}

	if(lr->lr_type == LRT_MASTER_LEAVING) {
		/*
		 * if we started this and (lr->u.lr_node_count != 0),
		 * then we have told everyone that we are leaving
		 */
		if ((lr->u.lr_starter == my_id) && lr->u.lr_node_count){
			lr->u.lr_coordinator = 0xDEAD;
			if(!(saddr->sin_addr.s_addr = nodeid_to_ipaddr(lr->u.lr_starter))){
				return -1;
			}
			saddr->sin_port = lr->u.lr_starter_port;
			return 0;
		}
	}

	/*
	 * Check if we have access to the log.  We may not
	 * yet have loaded this device.
	 */
	if (!lc) {
		lr->u.lr_node_count++;
		return 0;
	}

	if(lr->lr_type == LRT_MASTER_LEAVING){
		if (lr->u.lr_starter == lc->server_id)
			lc->server_id = 0xDEAD;
		lr->u.lr_node_count++;
		return 0;
	}

	/*
	 * New node joins and needs to know I am the server
	 * We shortcut the election here and respond directly
	 * to the inquirer
	 *
	if((lc->server_id == my_id) && !atomic_read(&lc->suspended)){
	*/
	if (lc->server_id == my_id) {
		int r = atomic_read(&lc->suspended);

		if (r >= 2) {
			DMDEBUG("I am the assigned server while suspended: %s",
				lc->uuid + (strlen(lc->uuid) - 8));
			lc->server_id = 0xDEAD;
			lr->u.lr_node_count++;
			return 0;
		}

		if (r == 1)
			DMDEBUG("I'm suspended, but still responding as server: %s",
				lc->uuid + (strlen(lc->uuid) - 8));
		lr->u.lr_coordinator = my_id;
		if(!(saddr->sin_addr.s_addr = nodeid_to_ipaddr(lr->u.lr_starter))){
			return -1;
		}
		saddr->sin_port = lr->u.lr_starter_port;
		return 0;
	}

	if(lr->lr_type == LRT_ELECTION){
		if((lr->u.lr_starter == my_id) && (lr->u.lr_node_count)){
			/*
			 * We started this election, and we've been
			 * around the loop.  If the node count hasn't
			 * changed since we started, we can proceed to
			 * selection.  Otherwise, go again setting
			 * ourself as the leader to start.
			 */
			if(node_count == lr->u.lr_node_count){
				lr->lr_type = LRT_SELECTION;
			} else {
				lr->u.lr_coordinator = my_id;
			}
			lr->u.lr_node_count = 1;
			return 0;
		}

		/*
		 * We are in the election phase, so
		 * if we have the lowest ID so far,
		 * we elect ourselves for server.
		 *
		 * However, if the mirror is being suspended
		 * (lc->suspended), then we leave the current
		 * coordinator in place.
		 *
		 * The client must not set lc->suspended until
		 * it has completed sending all requests.  That
		 * way, everyone is done sending requests when
		 * the last server is stuck holding the ball.
		 */
		lr->u.lr_node_count++;
		
		if((my_id < lr->u.lr_coordinator) && !atomic_read(&lc->suspended)){
			lr->u.lr_coordinator = my_id;
		}
		return 0;
	} else if(lr->lr_type == LRT_SELECTION){
		if(lr->u.lr_starter != my_id){
			lr->u.lr_node_count++;
			return 0;
		}
		
		/*
		 * Need to restart election if someone
		 * has joined since we started.
		 *
		 * Here, we are the started, so set
		 * node_count = 1
		 */
		if(lr->u.lr_node_count == node_count){
			lr->lr_type = LRT_MASTER_ASSIGN;
		} else {
			lr->lr_type = LRT_ELECTION;
			lr->u.lr_coordinator = my_id;
			return 0;
		}
		lr->u.lr_node_count = 1;
	} else if(lr->lr_type == LRT_MASTER_ASSIGN){
		/*
		 * If we are the server, assign it
		 */
		if(lr->u.lr_coordinator == my_id){
			lc->server_valid = 0;
			lc->server_id = my_id;
		}

		/*
		 * Continue around the loop
		 */
		if(lr->u.lr_starter != my_id){
			return 0;
		}

		/*
		 * If I was the one who asked for the election,
		 * the send the results back to the client
		 */
		if(!(saddr->sin_addr.s_addr = nodeid_to_ipaddr(lr->u.lr_starter))){
			return -1;
		}
		saddr->sin_port = lr->u.lr_starter_port;
		lc->server_id = lr->u.lr_coordinator;
	}
	return 0;
}


/**
 * process_log_request
 * @sock - the socket to receive requests on
 *
 * This function receives a region request for a specific
 * mirror set/region.  ATTENTION -- fill rest of desc.
 *
 * Returns: 0 on success, -1 on error
 */
static int process_log_request(struct socket *sock){
	static struct log_request *lr = NULL;
	int error;
	uint32_t nodeid;
	struct msghdr msg;
	struct iovec iov;
	struct sockaddr_in saddr_in;
	mm_segment_t fs;
	struct log_c *lc;

	if (unlikely(!lr))
		lr = kmalloc(sizeof(*lr), GFP_KERNEL);
	if (!lr)
		return -1;

	memset(lr, 0, sizeof(struct log_request));
	memset(&saddr_in, 0, sizeof(saddr_in));
		
	msg.msg_control = NULL;
	msg.msg_controllen = 0;
	msg.msg_iovlen = 1;
	msg.msg_iov = &iov;
	msg.msg_flags = 0;
	msg.msg_name = &saddr_in;
	msg.msg_namelen = sizeof(saddr_in);
	iov.iov_len = sizeof(struct log_request);
	iov.iov_base = lr;
		
	fs = get_fs();
	set_fs(get_ds());
		
	error = my_recvmsg(sock, &msg, sizeof(struct log_request),
			     0, 5);
	set_fs(fs);
		
	if(error > 0){
		if(error < sizeof(struct log_request)){
			DMERR("Cluster mirror log server received incomplete message.");
		}
		lc = get_log_context(lr->lr_uuid, lr->lr_uuid_ref);

		if(lr->lr_type == LRT_ELECTION ||
		   lr->lr_type == LRT_SELECTION ||
		   lr->lr_type == LRT_MASTER_ASSIGN ||
		   lr->lr_type == LRT_MASTER_LEAVING){
			uint32_t old = (lc)?lc->server_id: 0xDEAD;
			if(process_election(lr, lc, &saddr_in)){
				DMERR("Election processing failed.");
				return -1;
			}
			if(lc && (old != lc->server_id) && (my_id == lc->server_id)){
				DMDEBUG("I'm the cluster mirror log server for %s",
				       lc->uuid + (strlen(lc->uuid) - 8));
				if (atomic_read(&lc->suspended)) {
					DMDEBUG("Not reading disk log because I'm suspended (%s)",
						lc->uuid + (strlen(lc->uuid) - 8));
					
				} else if (disk_resume(lc) == -EDEADLK) {
					DMDEBUG("Unable to read disk log - deadlock potential.");
				}
			}
			goto reply;
		}

		if(!lc){
			lr->u.lr_int_rtn = -ENXIO;
			goto reply;
		}

		if (lc->server_id != my_id) {
			lr->u.lr_int_rtn = -ENXIO;
			goto reply;
		}

		if (atomic_read(&lc->suspended)) {
			nodeid = ipaddr_to_nodeid((struct sockaddr *)msg.msg_name);
			/*
			DMDEBUG("Getting request while server (%u) is suspended:", my_id);
			DMDEBUG(" - Requester :: %u", nodeid);
			DMDEBUG(" - log uuid  :: %s", lc->uuid + (strlen(lc->uuid) - 8));
			DMDEBUG(" - req type  :: %s", RQ_STRING(lr->lr_type));
			*/
			if (my_id != nodeid) {
				lr->u.lr_int_rtn = -ENXIO;
				goto reply;
			}
		}			

		switch(lr->lr_type){
		case LRT_IS_CLEAN:
			error = server_is_clean(lc, lr);
			break;
		case LRT_IS_REMOTE_RECOVERING:
			error = server_is_remote_recovering(lc, lr);
			break;
		case LRT_IN_SYNC:
			error = server_in_sync(lc, lr);
			break;
		case LRT_MARK_REGION:
			if(!(nodeid = 
			     ipaddr_to_nodeid((struct sockaddr *)msg.msg_name))){
				error = -ENXIO;
				break;
			}
			error = server_mark_region(lc, lr, nodeid);
			lr->u.lr_int_rtn = 0;
			break;
		case LRT_CLEAR_REGION:
			if(!(nodeid = 
			     ipaddr_to_nodeid((struct sockaddr *)msg.msg_name))){
				error = -ENXIO;
				break;
			}
			error = server_clear_region(lc, lr, nodeid);
			break;
		case LRT_FLUSH:
			if(!(nodeid = 
			     ipaddr_to_nodeid((struct sockaddr *)msg.msg_name))){
				error = -ENXIO;
				break;
			}
			error = server_flush(lc, nodeid);
			break;
		case LRT_GET_RESYNC_WORK:
			if(!(nodeid = 
			     ipaddr_to_nodeid((struct sockaddr *)msg.msg_name))){
				error = -ENXIO;
				break;
			}
			error = server_get_resync_work(lc, lr, nodeid);
			break;
		case LRT_COMPLETE_RESYNC_WORK:
			if(!(nodeid =
			     ipaddr_to_nodeid((struct sockaddr *)msg.msg_name))){
				error = -ENXIO;
				break;
			}
			error = server_complete_resync_work(lc, lr, lr->u.lr_int_rtn, nodeid);
			lr->u.lr_int_rtn = 0;
			break;
		case LRT_GET_SYNC_COUNT:
			error = server_get_sync_count(lc, lr);
			break;
		default:
			DMWARN("unknown request type received");
			return 0;  /* do not send a reply */
			break;
		}

		/* ATTENTION -- if error? */
		if(error){
/*
			DMWARN("Error (%d) while processing request (%s)",
			       error, RQ_STRING(lr->lr_type));
*/
			lr->u.lr_int_rtn = error;
		}
	reply:
    
		/* Why do we need to reset this? */
		iov.iov_len = sizeof(struct log_request);
		iov.iov_base = lr;
		msg.msg_name = &saddr_in;
		msg.msg_namelen = sizeof(saddr_in);

		fs = get_fs();
		set_fs(get_ds());
			
		error = sock_sendmsg(sock, &msg, sizeof(struct log_request));
			
		set_fs(fs);
		if(error < 0){
			DMWARN("unable to sendmsg to client (type = %s, error = %d)",
			       RQ_STRING(lr->lr_type), error);
			return error;
		}
	} else if(error == -EAGAIN || error == -ETIMEDOUT){
		return 0;
	} else {
		/* ATTENTION -- what do we do with this ? */
		return error;
	}
	return 0;
}

static int cluster_log_serverd(void *data){
	int error;
	struct log_c *lc;
	struct sockaddr_in saddr_in;
	struct socket *sock;

	/* read the disk logs */

	daemonize("cluster_log_serverd");

	error = sock_create(AF_INET, SOCK_DGRAM,
			    0,
			    &sock);
	if(error < 0){
		DMWARN("failed to create cluster mirror log server socket.");
		goto fail1;
	}

	memset(&saddr_in, 0, sizeof(struct sockaddr_cl));
	saddr_in.sin_family = AF_INET;
	saddr_in.sin_port = CLUSTER_LOG_PORT;
	error = sock->ops->bind(sock, (struct sockaddr *)&saddr_in,
				sizeof(saddr_in));

	if(error < 0){
		DMWARN("failed to bind cluster mirror log server socket.");
		goto fail2;
	}

	complete(&server_completion);
  
	DMDEBUG("cluster_log_serverd ready for work");
	for(;;){
		if(!atomic_read(&server_run)){
			break;
		}

		switch(atomic_read(&restart_event_type)){
		case SERVICE_NODE_LEAVE:
			/* ATTENTION -- may wish to check if regions **
			** are still in use by this node.  For now,  **
			** we do the same as if the node failed.  If **
			** there are no region still in-use by the   **
			** leaving node, it won't hurt anything - and**
			** if there is, they will be recovered.      */
		case SERVICE_NODE_FAILED:
			if (atomic_read(&restart_event_type) == SERVICE_NODE_FAILED)
				DMINFO("A cluster mirror log member has failed.");
			
			down(&log_list_lock);
			list_for_each_entry(lc, &log_list_head, log_list){
				if(lc->server_id == my_id){
					if (atomic_read(&lc->suspended)) {
						DMDEBUG("Not reading disk log because I'm suspended.");
					} else if (disk_resume(lc) == -EDEADLK) {
						DMDEBUG("Unable to read disk log - deadlock potential.");
					}
				}
			}
			up(&log_list_lock);

			break;
		case SERVICE_NODE_JOIN:
			DMDEBUG("Node joining");
			break;
		default:
			/* Someone has joined, or there is no event */
			break;
		}
		
		
		if(atomic_read(&restart_event_type)){
			/* finish the start phase */
			kcl_start_done(local_id, restart_event_id);
			restart_event_id = 0;

			/* Trigger any waiting starts to proceed */
			atomic_set(&restart_event_type, 0);
		} else if (atomic_read(&_do_requests)) {
			/* ATTENTION -- what to do with error ? */
			if(process_log_request(sock))
				DMINFO("process_log_request:: failed");
		}
		schedule();
	}

	DMDEBUG("Closing socket on server side");
	sock_release(sock);
	complete(&server_completion);
	return 0;

 fail2:
	sock_release(sock);
 fail1:
	DMWARN("Server thread failed to start");
	atomic_set(&server_run, 0);
	complete(&server_completion);
	return error;
}

/*
void print_server_status(struct log_c *lc){
	int i;

	DMINFO("SERVER OUTPUT::");

	DMINFO("  Live nodes        :: %d", global_count);
	DMINFO("  Sync count        :: %Lu", lc->sync_count);
	DMINFO("  Disk Region count :: %Lu", lc->header.nr_regions);
	DMINFO("  Region count      :: %Lu", lc->region_count);
	DMINFO("  nr_regions        :: %Lu", lc->header.nr_regions);
	DMINFO("  region_count      :: %Lu", lc->region_count);

	if(lc->header.nr_regions != lc->region_count){
		DMINFO("  NOTE:  Mapping has changed.");
	}

	DMINFO("Marked regions::");
	i = print_zero_bits((unsigned char *)lc->clean_bits, 0, lc->bitset_uint32_count);
	DMINFO("  Total = %d", i);

	DMINFO("Out-of-sync regions::");
	i = print_zero_bits((unsigned char *)lc->sync_bits, 0, lc->bitset_uint32_count);
	DMINFO("  Total = %d", i);

}
*/

int server_busy(struct log_c *lc)
{
	if (!list_empty(&lc->region_users) ||
	    (lc->recovering_region != (uint64_t)-1))
		return 1;
	else
		return 0;
}

int server_free_region_users(struct log_c *lc)
{
	int i = 0;
	struct region_user *ru, *tmp_ru;

	list_for_each_entry_safe(ru, tmp_ru, &lc->region_users, ru_list) {
		i++;
		list_del(&ru->ru_list);
		mempool_free(ru, region_user_pool);
	}

	DMDEBUG("%d region user structures freed", i);
	return 0;
}

int resume_server_requests(void) {
	atomic_set(&_do_requests, 1);
	return 0;
}

int start_server(void /* log_devices ? */){
	int error;

	DMDEBUG("start_server called");
	region_user_pool = mempool_create(1000, region_user_alloc,
					  region_user_free, NULL);
	if(!region_user_pool){
		DMWARN("unable to allocate region user pool for server");
		return -ENOMEM;
	}

	atomic_set(&_do_requests, 0);
	atomic_set(&server_run, 1);
	init_completion(&server_completion);

	error = kernel_thread(cluster_log_serverd, NULL, 0);
	if(error < 0){
		mempool_destroy(region_user_pool);
		DMWARN("failed to start kernel thread.");
		return error;
	}
	wait_for_completion(&server_completion);

	if(!atomic_read(&server_run)){
		mempool_destroy(region_user_pool);
		DMWARN("Cluster mirror log server thread failed to start");
		return -1;
	}
	dm_io_get(32);
	return 0;
}


void stop_server(void){
	DMDEBUG("stop_server called");
	atomic_set(&server_run, 0);

	wait_for_completion(&server_completion);
	down(&log_list_lock);
	if (!list_empty(&log_list_head)) {
		DMERR("Log elements remain at cluster log server shutdown");
	}
	up(&log_list_lock);
	mempool_destroy(region_user_pool);

	dm_io_put(32);
}
/*
 * Overrides for Emacs so that we follow Linus's tabbing style.
 * Emacs will notice this stuff at the end of the file and automatically
 * adjust the settings for this buffer only.  This must remain at the end
 * of the file.
 * ---------------------------------------------------------------------------
 * Local variables:
 * c-file-style: "linux"
 * End:
 */
